from jf_agent.session import retry_session
from requests.utils import default_user_agent
import requests

from collections import deque
from datetime import datetime
import time
import pytz
import logging

logger = logging.getLogger(__name__)


class GithubClient:
    def __init__(self, token, base_url=None, verify=True, session=None, **kwargs):
        self.token = token
        self.base_url = base_url or 'https://api.github.com'

        self.session = session or retry_session(**kwargs)
        self.session.verify = verify
        self.session.headers.update(
            {
                'Accept': 'application/json',
                'User-Agent': default_user_agent(),
                'Authorization': f'token {token}',
            }
        )

    def get_organization_by_name(self, org):
        url = f'{self.base_url}/orgs/{org}'
        return self.get_json(url)

    def get_all_users(self, org):
        url = f'{self.base_url}/orgs/{org}/members'
        return (self.get_json(m['url']) for m in self.get_all_pages(url))

    def get_all_repos(self, org):
        url = f'{self.base_url}/orgs/{org}/repos'
        result = (self.get_json(m['url']) for m in self.get_all_pages(url))
        return result

    def get_branches(self, full_repo):
        url = f'{self.base_url}/repos/{full_repo}/branches'
        return self.get_all_pages(url)

    def get_commits(self, full_repo, sha, since, until):
        url = f'{self.base_url}/repos/{full_repo}/commits?sha={sha}&since={since.isoformat()}&until={until.isoformat()}'
        return self.get_all_pages(url)

    def get_pullrequests(self, full_repo):
        url = f'{self.base_url}/repos/{full_repo}/pulls?state=all&sort=updated&direction=desc'
        return (self.get_json(m['url']) for m in self.get_all_pages(url))

    def get_pr_comments(self, full_repo, pr_id):
        url = f'{self.base_url}/repos/{full_repo}/pulls/{pr_id}/comments'
        return self.get_all_pages(url)

    def get_pr_reviews(self, full_repo, pr_id):
        url = f'{self.base_url}/repos/{full_repo}/pulls/{pr_id}/reviews'
        return self.get_all_pages(url)

    def get_pr_commits(self, full_repo, pr_id):
        url = f'{self.base_url}/repos/{full_repo}/pulls/{pr_id}/commits'
        return self.get_all_pages(url)

    # Raw web service operations with optional rate limiting
    def get_json(self, url):
        return self.get_raw_result(url).json()

    def get_raw_result(self, url):
        # retry if rate-limited
        max_retries = 3
        for i in range(1, max_retries + 1):
            try:
                result = self.session.get(url)
                result.raise_for_status()
                return result
            except requests.exceptions.HTTPError as e:
                if e.response.headers['X-RateLimit-Remaining'] == '0' and i < max_retries:
                    # rate-limited!  Sleep until it's oK, then try again
                    reset_time = datetime.fromtimestamp(
                        int(e.response.headers['X-RateLimit-Reset']), pytz.utc
                    )
                    now = datetime.utcnow().replace(tzinfo=pytz.utc)
                    reset_wait = reset_time - now
                    logger.warn(
                        f'Github rate limit exceeded.  Trying again in {str(reset_wait)}...'
                    )
                    time.sleep(reset_wait.seconds)
                    continue

                raise

    # Handle pagination
    def get_all_pages(self, url):
        current_page_values = deque()
        while True:
            # current page is exhausted; get a new page if there is one
            if not current_page_values:
                if not url:
                    return  # no next page

                # fetch the next page
                result = self.get_raw_result(url)
                page = result.json()
                if type(page) != list:
                    raise ValueError(f'Expected an array of json results, but got: {page}')

                if len(page) == 0:
                    return  # no new values returned

                current_page_values.extend(page)
                url = result.links['next']['url'] if 'next' in result.links else None

            yield current_page_values.popleft()



import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry


class ReauthSession(requests.Session):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def request(self, method, url, **kwargs):
        # If we get HTTP 401, re-authenticate and try again
        response = super().request(method, url, **kwargs)
        if response.status_code == 401:
            print(f'WARN: received 401 for the request [{method}] {url} - resetting client session')
            # Clear cookies and re-auth
            self.cookies.clear()
            response = super().request(method, url, **kwargs)
            self.cookies = response.cookies
        return response


def retry_session(**kwargs):
    """
    Obtains a requests session with retry settings.
    :return: session: Session
    """

    session = ReauthSession(**kwargs)

    retries = 3
    backoff_factor = 0.5
    status_forcelist = (500, 502, 504)

    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )

    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)

    return session



from datetime import datetime
import stashy
import re
import pytz
from tqdm import tqdm

from jf_agent.name_redactor import NameRedactor, sanitize_text

_branch_redactor = NameRedactor(preserve_names=['master', 'develop'])
_project_redactor = NameRedactor()
_repo_redactor = NameRedactor()


def datetime_from_bitbucket_server_timestamp(bb_server_timestamp_str):
    return datetime.fromtimestamp(float(bb_server_timestamp_str) / 1000).replace(tzinfo=pytz.utc)


def _normalize_user(user):
    return {
        'id': user['id'],
        'login': user['name'],
        'name': user['displayName'],
        'email': user.get('emailAddress', ''),
    }


def get_all_users(client):
    print('downloading bitbucket users... ', end='', flush=True)
    users = [_normalize_user(user) for user in client.admin.users]
    print('✓')
    return users


def _normalize_project(api_project, redact_names_and_urls):
    return {
        'id': api_project['id'],
        'login': api_project['key'],
        'name': (
            api_project.get('name')
            if not redact_names_and_urls
            else _project_redactor.redact_name(api_project.get('name'))
        ),
        'url': api_project['links']['self'][0]['href'] if not redact_names_and_urls else None,
    }


def get_all_projects(client, include_projects, exclude_projects, redact_names_and_urls):
    print('downloading bitbucket projects... ', end='', flush=True)

    filters = []
    if include_projects:
        filters.append(lambda p: p['key'] in include_projects)
    if exclude_projects:
        filters.append(lambda p: p['key'] not in exclude_projects)

    projects = [
        (p, _normalize_project(p, redact_names_and_urls))
        for p in client.projects.list()
        if all(filt(p) for filt in filters)
    ]
    print('✓')
    return projects


def _normalize_repo(api_project, api_repo, redact_names_and_urls):
    repo = api_repo.get()

    name = (
        repo['name'] if not redact_names_and_urls else _project_redactor.redact_name(repo['name'])
    )
    url = repo['links']['self'][0]['href'] if not redact_names_and_urls else None
    try:
        default_branch_name = (
            api_repo.default_branch['displayId'] if api_repo.default_branch else ''
        )
    except stashy.errors.NotFoundException:
        default_branch_name = ''

    branches = [
        {
            'name': (
                b['displayId']
                if not redact_names_and_urls
                else _branch_redactor.redact_name(b['displayId'])
            ),
            'sha': b['latestCommit'],
        }
        for b in api_repo.branches()
    ]

    return {
        'id': repo['id'],
        'name': name,
        'full_name': name,
        'url': url,
        'default_branch_name': default_branch_name,
        'branches': branches,
        'is_fork': 'origin' in repo,
        'project': _normalize_project(api_project, redact_names_and_urls),
    }


def get_all_repos(client, api_projects, include_repos, exclude_repos, redact_names_and_urls):
    print('downloading bitbucket repositories... ', end='', flush=True)

    filters = []
    if include_repos:
        filters.append(lambda r: r['name'] in include_repos)
    if exclude_repos:
        filters.append(lambda r: r['name'] not in exclude_repos)

    for api_project in api_projects:
        project = client.projects[api_project['key']]
        for repo in project.repos.list():
            if all(filt(repo) for filt in filters):
                api_repo = project.repos[repo['name']]
                yield api_repo, _normalize_repo(api_project, api_repo, redact_names_and_urls)

    print('✓')


def _normalize_commit(commit, repo, strip_text_content, redact_names_and_urls):
    return {
        'hash': commit['id'],
        'commit_date': datetime_from_bitbucket_server_timestamp(commit['committerTimestamp']),
        'author': commit['author'],
        'author_date': datetime_from_bitbucket_server_timestamp(commit['authorTimestamp']),
        'url': (
            repo['links']['self'][0]['href'].replace('browse', f'commits/{commit["id"]}')
            if not redact_names_and_urls
            else None
        ),
        'message': sanitize_text(commit.get('message'), strip_text_content),
        'is_merge': len(commit['parents']) > 1,
        'repo': _normalize_pr_repo(repo, redact_names_and_urls),
    }


def get_default_branch_commits(
    client, api_repos, strip_text_content, pull_since, pull_until, redact_names_and_urls
):
    for api_repo in api_repos:
        repo = api_repo.get()
        api_project = client.projects[repo['project']['key']]

        try:
            default_branch = api_repo.default_branch['displayId'] if api_repo.default_branch else ''
            commits = api_project.repos[repo['name']].commits(until=default_branch)

            for commit in tqdm(
                commits, desc=f'downloading commits for {repo["name"]}', unit='commit'
            ):
                normalized_commit = _normalize_commit(
                    commit, repo, strip_text_content, redact_names_and_urls
                )
                # commits are ordered newest to oldest; if this isn't
                # old enough, skip it and keep going
                if normalized_commit['commit_date'] >= pull_until:
                    continue

                # if this is too old, we're done with this repo
                if pull_since and normalized_commit['commit_date'] < pull_since:
                    break

                yield normalized_commit

        except stashy.errors.NotFoundException as e:
            print(
                f'WARN: Got NotFoundException for branch {repo["default_branch_name"]}: {e}. Skipping...'
            )


def _normalize_pr_repo(repo, redact_names_and_urls):
    normal_repo = {
        'id': repo['id'],
        'name': (
            repo['name'] if not redact_names_and_urls else _repo_redactor.redact_name(repo['name'])
        ),
    }

    if not redact_names_and_urls:
        if 'links' in repo:
            normal_repo['url'] = repo['links']['self'][0]['href']
        elif 'url' in repo:
            normal_repo['url'] = repo['url']

    return normal_repo


def get_pull_requests(
    client, api_repos, strip_text_content, pull_since, pull_until, redact_names_and_urls
):
    for api_repo in api_repos:
        repo = api_repo.get()
        api_project = client.projects[repo['project']['key']]
        api_repo = api_project.repos[repo['name']]

        for pr in tqdm(
            api_repo.pull_requests.all(state='ALL', order='NEWEST'),
            desc=f'downloading PRs for {repo["name"]}',
            unit='pr',
        ):
            updated_at = datetime_from_bitbucket_server_timestamp(pr['updatedDate'])
            # PRs are ordered newest to oldest; if this isn't old enough, skip it and keep going
            if updated_at >= pull_until:
                continue

            # if this is too old, we're done with this repo
            if pull_since and updated_at < pull_since:
                break

            api_pr = api_repo.pull_requests[pr['id']]

            try:
                pr_diffs = api_pr.diff().diffs
            except TypeError:
                additions, deletions, changed_files = None, None, None
            except stashy.errors.NotFoundException:
                additions, deletions, changed_files = None, None, None
            else:
                additions, deletions, changed_files = 0, 0, 0

                for pr_diff in pr_diffs:
                    changed_files += 1
                    for hunk in pr_diff.hunks:
                        for segment in hunk['segments']:
                            if segment['type'] == 'ADDED':
                                additions += len(segment['lines'])
                            if segment['type'] == 'REMOVED':
                                deletions += len(segment['lines'])

            comments = []
            approvals = []
            merge_date = None
            merged_by = None

            for activity in sorted(
                [a for a in api_pr.activities()], key=lambda x: x['createdDate']
            ):
                if activity['action'] == 'COMMENTED':
                    comments.append(
                        {
                            'user': _normalize_user(activity['comment']['author']),
                            'body': sanitize_text(activity['comment']['text'], strip_text_content),
                            'created_at': datetime_from_bitbucket_server_timestamp(
                                activity['comment']['createdDate']
                            ),
                        }
                    )
                elif activity['action'] in ('APPROVED', 'REVIEWED'):
                    approvals.append(
                        {
                            'foreign_id': activity['id'],
                            'user': _normalize_user(activity['user']),
                            'review_state': activity['action'],
                        }
                    )
                elif activity['action'] == 'MERGED':
                    merge_date = datetime_from_bitbucket_server_timestamp(activity['createdDate'])
                    merged_by = _normalize_user(activity['user'])

            closed_date = (
                datetime_from_bitbucket_server_timestamp(pr['closedDate'])
                if pr.get('closedDate')
                else None
            )

            try:
                commits = [
                    _normalize_commit(c, repo, strip_text_content, redact_names_and_urls)
                    for c in tqdm(
                        api_pr.commits(),
                        f'downloading commits for PR {pr["id"]}',
                        leave=False,
                        unit='commit',
                    )
                ]
            except stashy.errors.NotFoundException:
                print(
                    f'WARN: For PR {pr["id"]}, caught stashy.errors.NotFoundException when attempting to fetch a commit'
                )
                commits = []

            normalized_pr = {
                'id': pr['id'],
                'author': _normalize_user(pr['author']['user']),
                'title': sanitize_text(pr['title'], strip_text_content),
                'body': sanitize_text(pr.get('description'), strip_text_content),
                'is_closed': pr['state'] != 'OPEN',
                'is_merged': pr['state'] == 'MERGED',
                'created_at': datetime_from_bitbucket_server_timestamp(pr['createdDate']),
                'updated_at': updated_at,
                'closed_date': closed_date,
                'url': (pr['links']['self'][0]['href'] if not redact_names_and_urls else None),
                'base_repo': _normalize_pr_repo(pr['toRef']['repository'], redact_names_and_urls),
                'base_branch': (
                    pr['toRef']['displayId']
                    if not redact_names_and_urls
                    else _branch_redactor.redact_name(pr['toRef']['displayId'])
                ),
                'head_repo': _normalize_pr_repo(pr['fromRef']['repository'], redact_names_and_urls),
                'head_branch': (
                    pr['fromRef']['displayId']
                    if not redact_names_and_urls
                    else _branch_redactor.redact_name(pr['fromRef']['displayId'])
                ),
                'additions': additions,
                'deletions': deletions,
                'changed_files': changed_files,
                'comments': comments,
                'approvals': approvals,
                'merge_date': merge_date,
                'merged_by': merged_by,
                'commits': commits,
            }

            yield normalized_pr



from tqdm import tqdm
import queue
import threading
import sys
import string
import json
import logging

from jira.resources import dict2resource
from jira.exceptions import JIRAError
from jira.utils import json_loads


logger = logging.getLogger(__name__)


# Returns an array of User dicts
def download_users(jira_connection, gdpr_active):
    print('downloading jira users... ', end='', flush=True)

    jira_users = _search_all_users(jira_connection, gdpr_active)

    # Some jira instances won't return more than one page of
    # results.  If we have seen exactly 1000 results, try
    # searching a different way
    if len(jira_users) == 1000:
        jira_users = _users_by_letter(jira_connection, gdpr_active)

    print('✓')
    return jira_users


# Returns an array of Field dicts
def download_fields(jira_connection, include_fields, exclude_fields):

    print('downloading jira fields... ', end='', flush=True)

    filters = []
    if include_fields:
        filters.append(lambda field: field['id'] in include_fields)
    if exclude_fields:
        filters.append(lambda field: field['id'] not in exclude_fields)

    fields = [field for field in jira_connection.fields() if all(filt(field) for filt in filters)]

    print('✓')
    return fields


# Returns an array of Resolutions dicts
def download_resolutions(jira_connection):
    print('downloading jira resolutions... ', end='', flush=True)
    result = [r.raw for r in jira_connection.resolutions()]
    print('✓')
    return result


# Returns an array of IssueType dicts
def download_issuetypes(jira_connection, project_ids):
    '''
    For Jira next-gen projects, issue types can be scoped to projects.
    For issue types that are scoped to projects, only extract the ones
    in the extracted projects.
    '''
    print('downloading jira issue types... ', end='', flush=True)
    result = []
    for it in jira_connection.issue_types():
        if 'scope' in it.raw and it.raw['scope']['type'] == 'PROJECT':
            if it.raw['scope']['project']['id'] in project_ids:
                result.append(it.raw)
        else:
            result.append(it.raw)
    print('✓')
    return result


# Returns an array of LinkType dicts
def download_issuelinktypes(jira_connection):
    print('downloading jira issue link types... ', end='', flush=True)
    result = [lt.raw for lt in jira_connection.issue_link_types()]
    print('✓')
    return result


# Returns an array of Priority dicts
def download_priorities(jira_connection):
    print('downloading jira priorities... ', end='', flush=True)
    result = [p.raw for p in jira_connection.priorities()]
    print('✓')
    return result


# Each project has a list of versions.
# Returns an array of Project dicts, where each one is agumented with an array of associated Version dicts.
def download_projects_and_versions(
    jira_connection, include_projects, exclude_projects, include_categories, exclude_categories
):

    print('downloading jira projects... ', end='', flush=True)

    filters = []
    if include_projects:
        filters.append(lambda proj: proj.key in include_projects)
    if exclude_projects:
        filters.append(lambda proj: proj.key not in exclude_projects)
    if include_categories:
        filters.append(lambda proj: proj.projectCategory.name in include_categories)
    if exclude_categories:
        filters.append(lambda proj: proj.projectCategory.name not in exclude_categories)

    projects = [proj for proj in jira_connection.projects() if all(filt(proj) for filt in filters)]

    print('✓')

    print('downloading jira versions... ', end='', flush=True)
    result = [
        p.raw.update({'versions': [v.raw for v in jira_connection.project_versions(p)]}) or p.raw
        for p in projects
    ]
    print('✓')
    return result


# Boards and Sprints are many-to-many.
# Returns a 3-tuple:
#   - Array of board dicts
#   - Array of sprint dicts
#   - Array of board/sprint links
def download_boards_and_sprints(jira_connection, project_ids):
    b_start_at = 0
    boards = []
    print('downloading jira boards... ', end='', flush=True)
    while True:
        # boards seem to come back in batches of 50 at most
        jira_boards = jira_connection.boards(startAt=b_start_at, maxResults=50)
        if not jira_boards:
            break
        b_start_at += len(jira_boards)

        # Some versions of Jira map boards to projects, some do not.
        # If this includes a "location" for boards, then only
        # include boards for the projects we're pulling
        boards.extend(
            [
                b
                for b in jira_boards
                if not hasattr(b, 'location')
                or str(getattr(b.location, 'projectId', '')) in project_ids
            ]
        )
    print('✓')

    links = []
    sprints = {}
    for b in tqdm(boards, desc='downloading jira sprints', file=sys.stdout):
        if b.raw['type'] != 'scrum':
            continue
        s_start_at = 0
        sprints_for_board = []
        while True:
            batch = None
            try:
                batch = jira_connection.sprints(board_id=b.id, startAt=s_start_at, maxResults=50)
            except JIRAError as e:
                # JIRA returns 500 errors for various reasons: board is
                # misconfigured; "falied to execute search"; etc.  Just
                # skip and move on
                if e.status_code == 500 or e.status_code == 404:
                    logger.warn(f"Couldn't get sprints for board {b.id}.  Skipping...")
                else:
                    raise

            if not batch:
                break
            s_start_at += len(batch)
            sprints_for_board.extend(batch)

        links.append({'board_id': b.id, 'sprint_ids': [s.id for s in sprints_for_board]})
        sprints.update({s.id: s for s in sprints_for_board})

    return [b.raw for b in boards], [s.raw for s in sprints.values()], links


def download_issues(jira_connection, project_ids, include_fields, exclude_fields):
    issue_jql = f'created is not empty and project in ({",".join(project_ids)}) order by id asc'
    issue_count = jira_connection.search_issues(issue_jql, fields='id', maxResults=1).total
    parallel_threads = 10
    issues_per_thread = -(-issue_count // parallel_threads)

    # Make threads to talk to Jira and write batches of issues to the queue
    q = queue.Queue()
    threads = [
        threading.Thread(
            target=_download_some_jira_issues,
            args=[
                i,
                issue_jql,
                include_fields,
                exclude_fields,
                jira_connection,
                i * issues_per_thread,
                (i + 1) * issues_per_thread,
                q,
            ],
        )
        for i in range(parallel_threads)
    ]
    for t in threads:
        t.start()

    issues = {}
    with tqdm(desc='downloading jira issues', total=issue_count, file=sys.stdout) as prog_bar:
        # Read batches from queue
        finished = 0
        while finished < len(threads):
            batch = q.get()
            if batch is None:
                # found the marker that a thread is done
                finished += 1
                continue

            if isinstance(batch, BaseException):
                # thread had a problem; rethrow
                raise batch

            # issues sometimes appear at the end of one batch and again at the beginning of the next; use
            # the dict to dedupe
            old_count = len(issues)
            issues.update({i['id']: i for i in batch})
            new_count = len(issues) - old_count

            prog_bar.update(new_count)

    for t in threads:
        t.join()

    return list(issues.values())


# Returns a dict with two items: 'existing' gives a list of all worklogs
# that currently exist; 'deleted' gives the list of worklogs that
# existed at some point previously, but have since been deleted
def download_worklogs(jira_connection, issue_ids):
    print(f'downloading jira worklogs... ', end='', flush=True)
    updated = []
    since = 0
    while True:
        worklog_ids_json = jira_connection._get_json('worklog/updated', params={'since': since})
        updated_worklog_ids = [v['worklogId'] for v in worklog_ids_json['values']]

        resp = jira_connection._session.post(
            url=jira_connection._get_url('worklog/list'),
            data=json.dumps({'ids': updated_worklog_ids}),
        )
        try:
            worklog_list_json = json_loads(resp)
        except ValueError as e:
            logger.exception("Couldn't parse JIRA response as JSON: %s", resp.text)
            raise e

        updated.extend([wl for wl in worklog_list_json if wl['issueId'] in issue_ids])
        if worklog_ids_json['lastPage']:
            break
        since = worklog_ids_json['until']

    print('✓')

    print(f'Finding old worklogs that have been deleted... ', end='', flush=True)
    deleted = []
    while True:
        worklog_ids_json = jira_connection._get_json('worklog/deleted', params={'since': 0})
        deleted.extend(worklog_ids_json['values'])
        if worklog_ids_json['lastPage']:
            break

    print('✓')

    return {'existing': updated, 'deleted': deleted}


def _jira_user_key(user_dict, gdpr_active):
    if gdpr_active:
        return user_dict['accountId']
    else:
        return user_dict['key']


def _search_all_users(jira_connection, gdpr_active):
    jira_users = {}
    start_at = 0
    while True:
        # Search for '%' or '@' so we get results from cloud or
        # on-prem, the latter being what on-prem appears to use as
        # a wildcard
        users = _search_users(
            jira_connection, gdpr_active, query='%', start_at=start_at, include_inactive=True
        ) or _search_users(
            jira_connection, gdpr_active, query='@', start_at=start_at, include_inactive=True
        )
        if not users:
            return list(jira_users.values())

        jira_users.update({_jira_user_key(u, gdpr_active): u for u in users})

        start_at += len(users)


def _users_by_letter(jira_connection, gdpr_active):
    jira_users = {}
    for letter in string.ascii_lowercase:
        jira_users.update(
            {
                _jira_user_key(u, gdpr_active): u
                for u in _search_users(
                    jira_connection,
                    gdpr_active,
                    query=f'{letter}.',
                    include_inactive=True,
                    include_active=False,
                )
            }
        )
        jira_users.update(
            {
                _jira_user_key(u, gdpr_active): u
                for u in _search_users(
                    jira_connection,
                    gdpr_active,
                    query=f'{letter}.',
                    include_inactive=False,
                    include_active=True,
                )
            }
        )
    return list(jira_users.values())


def _search_users(
    jira_connection,
    gdpr_active,
    query,
    start_at=0,
    max_results=1000,
    include_active=True,
    include_inactive=False,
):
    if gdpr_active:
        # jira_connection.search_users creates a query that is no longer accepted on
        # GDPR-compliant Jira instances.  Construct the right query by hand
        params = {
            'startAt': start_at,
            'maxResults': max_results,
            'query': query,
            'includeActive': include_active,
            'includeInactive': include_inactive,
        }
        return jira_connection._get_json('user/search', params)

    return [
        u.raw
        for u in jira_connection.search_users(
            query,
            startAt=start_at,
            maxResults=max_results,
            includeInactive=include_inactive,
            includeActive=include_active,
        )
    ]


def _download_some_jira_issues(
    i, issue_jql, include_fields, exclude_fields, jira_connection, start_at, end_at, q
):
    # Empirically, Jira cloud won't return more than 100 issues at a
    # time even if you ask for more, but on-premise can do 1000
    #
    # This could be brittle, either with changes in Jira cloud, or with
    # different on-premise servers configured differently
    batch_size = 1000 if jira_connection.deploymentType == 'Server' else 100
    total_issues = end_at

    try:
        # pull batches and bulk load each into our database
        while start_at < min(end_at, total_issues):
            issues = _get_jira_issues_batch(
                issue_jql, include_fields, exclude_fields, jira_connection, start_at, batch_size
            )

            total_issues = issues.total
            issues_retrieved = len(issues)
            start_at += issues_retrieved
            if issues_retrieved == 0:
                logger.warn(
                    f'[Thread {i}] Jira issue downloaded might be stuck; asked for {batch_size} issues starting at {start_at}, '
                    f'ending at ending at {min(end_at, total_issues)}, got none back.  Trying again...'
                )

            raw_issues = [i.raw for i in issues]

            # TODO: configurable way to scrub things out of raw_issues here before we write them out.

            q.put(raw_issues)

        # sentinel to mark that this thread finished
        q.put(None)

    except BaseException as e:
        logger.info(f'[Thread {i}] Jira issue downloader FAILED')
        q.put(e)


def _get_jira_issues_batch(
    issue_jql, include_fields, exclude_fields, jira_connection, start_at, max_results
):
    get_changelog = True

    fieldspec = ','.join(include_fields) if include_fields else '*all'
    for f in exclude_fields:
        fieldspec += f',-{f}'

    while max_results > 0:
        search_kwargs = {
            'fields': fieldspec,
            'expand': 'renderedFields',
            'startAt': start_at,
            'maxResults': max_results,
        }
        if get_changelog:
            search_kwargs['expand'] += ',changelog'

        try:
            return _expand_changelog(
                jira_connection.search_issues(issue_jql, **search_kwargs), jira_connection
            )
        except JIRAError:
            logger.exception(f'JIRAError, reducing batch size')
            ## back off the max results requested to try and avoid the error
            max_results = int(max_results / 2)
            if max_results == 0:
                ## if we hit 0, then there's an issue we can't request with a changelog, fetch
                ## it without a change log and continue
                get_changelog = False
                max_results = 1


# Sometimes the results of issue search has an incomplete changelog.  Fill it in if so.
def _expand_changelog(jira_issues, jira_connection):
    for i in jira_issues:
        changelog = getattr(i, 'changelog', None)
        if changelog and changelog.total != changelog.maxResults:
            # expand the changelog
            start_at = changelog.maxResults
            batch_size = 100
            while start_at < changelog.total:
                more_cls = jira_connection._get_json(
                    f'issue/{i.id}/changelog', {'startAt': start_at, 'maxResults': batch_size}
                )['values']
                changelog.histories.extend(dict2resource(i) for i in more_cls)
                i.raw['changelog']['histories'].extend(more_cls)
                start_at += len(more_cls)
    return jira_issues



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="jf_agent",
    version="0.0.20",
    description="An agent for collecting data for jellyfish",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Jellyfish-AI/jf_agent",
    packages=setuptools.find_packages(),
    install_requires=["jira", "tqdm", "stashy", "pyyaml>=4.2b1", "urllib3==1.24.2", "dateparser"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={"console_scripts": ["jf_agent=jf_agent.main:main"]},
)



import argparse
import os
import sys
import json
import urllib3
import yaml
import dateparser
import pytz
from datetime import datetime
from stashy.client import Stash
from jf_agent.github_client import GithubClient
import logging

from jira import JIRA
from jira.resources import GreenHopperResource

from jf_agent.bb_download import (
    get_all_users as get_bb_users,
    get_all_projects as get_bb_projects,
    get_all_repos as get_bb_repos,
    get_default_branch_commits as get_bb_default_branch_commits,
    get_pull_requests as get_bb_pull_requests,
)

from jf_agent.gh_download import (
    get_all_users as get_gh_users,
    get_all_projects as get_gh_projects,
    get_all_repos as get_gh_repos,
    get_default_branch_commits as get_gh_default_branch_commits,
    get_pull_requests as get_gh_pull_requests,
)

from jf_agent.session import retry_session


from jf_agent.jira_download import (
    download_users,
    download_fields,
    download_resolutions,
    download_issuetypes,
    download_issuelinktypes,
    download_priorities,
    download_projects_and_versions,
    download_boards_and_sprints,
    download_issues,
    download_worklogs,
)


def main():
    logging.basicConfig(level=logging.WARNING)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-c', '--config-file', nargs='?', default='jellyfish.yaml', help='Path to config file'
    )
    parser.add_argument(
        '-s',
        '--since',
        nargs='?',
        default=None,
        help='Pull git activity on or after this timestamp',
    )
    parser.add_argument(
        '-u', '--until', nargs='?', default=None, help='Pull git activity before this timestamp'
    )

    args = parser.parse_args()

    with open(args.config_file, 'r') as ymlfile:
        config = yaml.safe_load(ymlfile)

    conf_global = config.get('global', {})
    outdir = conf_global.get('out_dir', '.')
    skip_ssl_verification = conf_global.get('no_verify_ssl', False)

    jira_config = config.get('jira', {})
    git_config = config.get('git', config.get('bitbucket', {}))

    jira_url = jira_config.get('url', None)

    git_url = git_config.get('url', None)

    pull_since = dateparser.parse(args.since) if args.since else None
    if pull_since:
        pull_since = pull_since.replace(tzinfo=pytz.utc)

    pull_until = dateparser.parse(args.until) if args.until else datetime.utcnow()
    pull_until = pull_until.replace(tzinfo=pytz.utc)

    if not jira_url and not git_url:
        print('ERROR: Config file must provide either a Jira or Bitbucket URL.')
        parser.print_help()
        sys.exit(1)

    if skip_ssl_verification:
        print('WARNING: Disabling SSL certificate validation')
        # To silence "Unverified HTTPS request is being made."
        urllib3.disable_warnings()

    if jira_url:
        jira_username = os.environ.get('JIRA_USERNAME', None)
        jira_password = os.environ.get('JIRA_PASSWORD', None)
        if jira_username and jira_password:
            jira_connection = get_basic_jira_connection(
                jira_url, jira_username, jira_password, skip_ssl_verification
            )
            if jira_connection:
                load_and_dump_jira(outdir, jira_config, jira_connection)
        else:
            print(
                'ERROR: Jira credentials not found. Set environment variables JIRA_USERNAME and JIRA_PASSWORD. Skipping Jira...'
            )

    if git_url:
        provider = git_config.get('provider', 'bitbucket_server')
        if provider not in ('bitbucket_server', 'github'):
            print(
                f'ERROR: unsupported git provider {provider}. Provider should be one of `bitbucket_server` or `github`'
            )
            return

        git_conn = get_git_client(provider, git_url, skip_ssl_verification)
        if git_conn:
            try:
                include_projects = set(git_config.get('include_projects', []))
                exclude_projects = set(git_config.get('exclude_projects', []))
                include_repos = set(git_config.get('include_repos', []))
                exclude_repos = set(git_config.get('exclude_projects', []))
                strip_text_content = git_config.get('strip_text_content', False)
                redact_names_and_urls = git_config.get('redact_names_and_urls', False)

                if provider == 'bitbucket_server':
                    load_and_dump_bb(
                        outdir,
                        git_conn,
                        pull_since,
                        pull_until,
                        include_projects,
                        exclude_projects,
                        include_repos,
                        exclude_repos,
                        strip_text_content,
                        redact_names_and_urls,
                    )
                elif provider == 'github':
                    load_and_dump_github(
                        outdir,
                        git_conn,
                        pull_since,
                        pull_until,
                        include_projects,
                        exclude_projects,
                        include_repos,
                        exclude_repos,
                        strip_text_content,
                        redact_names_and_urls,
                    )
                print()
                print(f'Pulled until: {pull_until}')
            except Exception as e:
                print(f'ERROR: Failed to download {provider} data:\n{e}')


def load_and_dump_jira(outdir, jira_config, jira_connection):
    try:
        include_fields = set(jira_config.get('include_fields', []))
        exclude_fields = set(jira_config.get('exclude_fields', []))

        fields = download_fields(jira_connection, include_fields, exclude_fields)

        print_fields_only = jira_config.get('print_fields_only', False)
        if print_fields_only:
            for f in fields:
                print(f"{f['key']:30}\t{f['name']}")
                return

        gdpr_active = jira_config.get('gdpr_active', False)

        include_projects = set(jira_config.get('include_projects', []))
        exclude_projects = set(jira_config.get('exclude_projects', []))
        include_categories = set(jira_config.get('include_project_categories', []))
        exclude_categories = set(jira_config.get('exclude_project_categories', []))

        write_file(outdir, 'jira_fields', fields)

        projects_and_versions = download_projects_and_versions(
            jira_connection,
            include_projects,
            exclude_projects,
            include_categories,
            exclude_categories,
        )
        project_ids = set([proj['id'] for proj in projects_and_versions])
        write_file(outdir, 'jira_projects_and_versions', projects_and_versions)

        write_file(outdir, 'jira_users', download_users(jira_connection, gdpr_active))
        write_file(outdir, 'jira_resolutions', download_resolutions(jira_connection))
        write_file(outdir, 'jira_issuetypes', download_issuetypes(jira_connection, project_ids))
        write_file(outdir, 'jira_linktypes', download_issuelinktypes(jira_connection))
        write_file(outdir, 'jira_priorities', download_priorities(jira_connection))

        boards, sprints, links = download_boards_and_sprints(jira_connection, project_ids)
        write_file(outdir, 'jira_boards', boards)
        write_file(outdir, 'jira_sprints', sprints)
        write_file(outdir, 'jira_board_sprint_links', links)

        issues = download_issues(jira_connection, project_ids, include_fields, exclude_fields)
        issue_ids = set([i['id'] for i in issues])
        write_file(outdir, 'jira_issues', issues)
        write_file(outdir, 'jira_worklogs', download_worklogs(jira_connection, issue_ids))
    except Exception as e:
        print(f'ERROR: Failed to download jira data:\n{e}')


def get_basic_jira_connection(url, username, password, skip_ssl_verification):
    try:
        return JIRA(
            server=url,
            basic_auth=(username, password),
            max_retries=3,
            options={
                'agile_rest_path': GreenHopperResource.AGILE_BASE_REST_PATH,
                'verify': not skip_ssl_verification,
            },
        )
    except Exception as e:
        print(f'ERROR: Failed to connect to Jira:\n{e}')


def load_and_dump_github(
    outdir,
    git_conn,
    pull_since,
    pull_until,
    include_projects,
    exclude_projects,
    include_repos,
    exclude_repos,
    strip_text_content,
    redact_names_and_urls,
):
    # github must be in whitelist mode
    if exclude_projects or not include_projects:
        print(
            'ERROR: Github cloud requires a list of projects (ie Github organizations) to pull from. Make sure you set for `include_projects` and not `exclude_projects`, and try again.'
        )
        return

    users = get_gh_users(git_conn, include_projects)
    write_file(outdir, 'bb_users', users)

    projects = get_gh_projects(git_conn, include_projects, redact_names_and_urls)
    write_file(outdir, 'bb_projects', projects)

    # turn a generator that produces (api_object, dict) pairs into separate lists of API objects and dicts
    api_repos, repos = zip(
        *get_gh_repos(
            git_conn, include_projects, include_repos, exclude_repos, redact_names_and_urls
        )
    )
    write_file(outdir, 'bb_repos', repos)

    commits = get_gh_default_branch_commits(
        git_conn, api_repos, strip_text_content, pull_since, pull_until, redact_names_and_urls
    )
    write_file(outdir, 'bb_commits', commits)

    prs = get_gh_pull_requests(
        git_conn, api_repos, strip_text_content, pull_since, pull_until, redact_names_and_urls
    )
    write_file(outdir, 'bb_prs', prs)


def load_and_dump_bb(
    outdir,
    bb_conn,
    pull_since,
    pull_until,
    include_projects,
    exclude_projects,
    include_repos,
    exclude_repos,
    strip_text_content,
    redact_names_and_urls,
):
    users = get_bb_users(bb_conn)
    write_file(outdir, 'bb_users', users)

    # turn a generator that produces (api_object, dict) pairs into separate lists of API objects and dicts
    api_projects, projects = zip(
        *get_bb_projects(bb_conn, include_projects, exclude_projects, redact_names_and_urls)
    )
    write_file(outdir, 'bb_projects', projects)

    # turn a generator that produces (api_object, dict) pairs into separate lists of API objects and dicts
    api_repos, repos = zip(
        *get_bb_repos(bb_conn, api_projects, include_repos, exclude_repos, redact_names_and_urls)
    )
    write_file(outdir, 'bb_repos', repos)

    commits = get_bb_default_branch_commits(
        bb_conn, api_repos, strip_text_content, pull_since, pull_until, redact_names_and_urls
    )
    write_file(outdir, 'bb_commits', commits)

    prs = get_bb_pull_requests(
        bb_conn, api_repos, strip_text_content, pull_since, pull_until, redact_names_and_urls
    )
    write_file(outdir, 'bb_prs', prs)


def get_git_client(provider, git_url, skip_ssl_verification):
    if provider == 'bitbucket_server':
        bb_username = os.environ.get('BITBUCKET_USERNAME', None)
        bb_password = os.environ.get('BITBUCKET_PASSWORD', None)
        if not bb_username or not bb_password:
            print(
                'ERROR: Bitbucket credentials not found. Set environment variables BITBUCKET_USERNAME and BITBUCKET_PASSWORD. Skipping Bitbucket...'
            )
            return

        try:
            return Stash(
                base_url=git_url,
                username=bb_username,
                password=bb_password,
                verify=not skip_ssl_verification,
                session=retry_session(),
            )
        except Exception as e:
            print(f'ERROR: Failed to connect to Bitbucket:\n{e}')
            return

    elif provider == 'github':
        gh_token = os.environ.get('GITHUB_TOKEN', None)
        if not gh_token:
            print(
                'ERROR: Github credentials not found. Set environment variable GITHUB_TOKEN. Skipping Github...'
            )
            return

        try:
            return GithubClient(
                base_url=git_url,
                token=gh_token,
                verify=not skip_ssl_verification,
                session=retry_session(),
            )

        except Exception as e:
            print(f'ERROR: Failed to connect to Github:\n{e}')
            return

    raise ValueError(f'unsupported git provider {provider}')


def write_file(outdir, name, results):
    with open(f'{outdir}/{name}.json', 'w') as outfile:
        json.dump(list(results), outfile, indent=2, default=str)


if __name__ == '__main__':
    main()



from tqdm import tqdm
from github import GithubException
from dateutil import parser

from jf_agent.name_redactor import NameRedactor, sanitize_text


_branch_redactor = NameRedactor(preserve_names=['master', 'develop'])
_project_redactor = NameRedactor()
_repo_redactor = NameRedactor()


def _normalize_user(user):
    if not user:
        return None

    # raw user, just have email (e.g. from a commit)
    if 'id' not in user:
        return {
            'id': user['email'],
            'login': user['email'],
            'name': user['name'],
            'email': user['email'],
        }

    # API user, where github matched to a known account
    return {'id': user['id'], 'login': user['login'], 'name': user['name'], 'email': user['email']}


def get_all_users(client, include_orgs):
    print('downloading github users... ', end='', flush=True)
    users = [_normalize_user(user) for org in include_orgs for user in client.get_all_users(org)]
    print('✓')

    if not users:
        raise ValueError('No users found.  Make sure your token has appropriate access to Github.')

    return users


def _normalize_project(api_org, redact_names_and_urls):
    return {
        'id': api_org['id'],
        'login': api_org['login'],
        'name': (
            api_org['name']
            if not redact_names_and_urls
            else _project_redactor.redact_name(api_org['name'])
        ),
        'url': api_org['html_url'] if not redact_names_and_urls else None,
    }


def get_all_projects(client, include_orgs, redact_names_and_urls):
    print('downloading github projects... ', end='', flush=True)
    projects = [
        _normalize_project(client.get_organization_by_name(org), redact_names_and_urls)
        for org in include_orgs
    ]
    print('✓')

    if not projects:
        raise ValueError(
            'No projects found.  Make sure your token has appropriate access to Github.'
        )
    return projects


def _normalize_repo(client, org_name, repo, redact_names_and_urls):
    return {
        'id': repo['id'],
        'name': (
            repo['name']
            if not redact_names_and_urls
            else _project_redactor.redact_name(repo['name'])
        ),
        'full_name': (
            repo['full_name']
            if not redact_names_and_urls
            else _project_redactor.redact_name(repo['full_name'])
        ),
        'url': repo['html_url'] if not redact_names_and_urls else None,
        'default_branch_name': repo['default_branch'],
        'branches': [
            {
                'name': (
                    b['name']
                    if not redact_names_and_urls
                    else _branch_redactor.redact_name(b['name'])
                ),
                'sha': b['commit']['sha'],
            }
            for b in client.get_branches(repo['full_name'])
        ],
        'is_fork': repo['fork'],
        'project': _normalize_project(
            client.get_json(repo['organization']['url']), redact_names_and_urls
        ),
    }


def get_all_repos(client, include_orgs, include_repos, exclude_repos, redact_names_and_urls):
    print('downloading github repos... ', end='', flush=True)

    filters = []
    if include_repos:
        filters.append(lambda r: r['name'] in include_repos)
    if exclude_repos:
        filters.append(lambda r: r['name'] not in exclude_repos)

    repos = [
        (r, _normalize_repo(client, org, r, redact_names_and_urls))
        for org in include_orgs
        for r in client.get_all_repos(org)
        if all(filt(r) for filt in filters)
    ]
    print('✓')
    if not repos:
        raise ValueError(
            'No repos found. Make sure your token has appropriate access to Github and check your configuration of repos to pull.'
        )
    return repos


def _normalize_commit(commit, repo, strip_text_content, redact_names_and_urls):
    author = commit.get('author') or {}
    author.update(
        {'name': commit['commit']['author']['name'], 'email': commit['commit']['author']['email']}
    )

    return {
        'hash': commit['sha'],
        'commit_date': commit['commit']['committer']['date'],
        'author': _normalize_user(author),
        'author_date': commit['commit']['author']['date'],
        'url': commit['html_url'] if not redact_names_and_urls else None,
        'message': sanitize_text(commit['commit']['message'], strip_text_content),
        'is_merge': len(commit['parents']) > 1,
        'repo': _normalize_pr_repo(repo, redact_names_and_urls),
    }


def _normalize_pr_repo(repo, redact_names_and_urls):
    return {
        'id': repo['id'],
        'name': (
            repo['name'] if not redact_names_and_urls else _repo_redactor.redact_name(repo['name'])
        ),
        'url': repo['html_url'] if not redact_names_and_urls else None,
    }


def get_default_branch_commits(
    client, api_repos, strip_text_content, pull_since, pull_until, redact_names_and_urls
):
    for repo in api_repos:
        try:
            for commit in tqdm(
                client.get_commits(
                    repo['full_name'], repo['default_branch'], since=pull_since, until=pull_until
                ),
                desc=f'downloading commits for {repo["name"]}',
                unit='commit',
            ):
                yield _normalize_commit(commit, repo, strip_text_content, redact_names_and_urls)

        except GithubException as e:
            print(f':WARN: Got exception for branch {repo["default_branch"]}: {e}. Skipping...')


def _normalize_pr(client, pr, strip_text_content, redact_names_and_urls):
    return {
        'id': pr['number'],
        'author': _normalize_user(client.get_json(pr['user']['url'])),
        'title': sanitize_text(pr['title'], strip_text_content),
        'body': sanitize_text(pr['body'], strip_text_content),
        'is_closed': pr['state'] == 'closed',
        'is_merged': pr['merged'],
        'created_at': pr['created_at'],
        'updated_at': pr['updated_at'],
        'closed_date': pr['closed_at'] if pr['closed_at'] else None,
        'url': (pr['html_url'] if not redact_names_and_urls else None),
        'base_repo': _normalize_pr_repo(pr['base']['repo'], redact_names_and_urls),
        'base_branch': (
            pr['base']['ref']
            if not redact_names_and_urls
            else _branch_redactor.redact_name(pr['base']['ref'])
        ),
        'head_repo': _normalize_pr_repo(pr['head']['repo'], redact_names_and_urls),
        'head_branch': (
            pr['head']['ref']
            if not redact_names_and_urls
            else _branch_redactor.redact_name(pr['head']['ref'])
        ),
        'additions': pr['additions'],
        'deletions': pr['deletions'],
        'changed_files': pr['changed_files'],
        'comments': [
            {
                'user': _normalize_user(client.get_json(c['user']['url'])),
                'body': c['body'],
                'created_at': c['created_at'],
            }
            for c in client.get_pr_comments(pr['base']['repo']['full_name'], pr['number'])
        ],
        'approvals': [
            {
                'foreign_id': r['id'],
                'user': _normalize_user(client.get_json(r['user']['url'])),
                'review_state': r['state'],
            }
            for r in client.get_pr_reviews(pr['base']['repo']['full_name'], pr['number'])
        ],
        'merge_date': pr['merged_at'] if pr['merged_at'] else None,
        'merged_by': (
            _normalize_user(client.get_json(pr['merged_by']['url'])) if pr['merged'] else None
        ),
        'commits': [
            _normalize_commit(c, pr['base']['repo'], strip_text_content, redact_names_and_urls)
            for c in tqdm(
                client.get_pr_commits(pr['base']['repo']['full_name'], pr['number']),
                f'downloading commits for PR {pr["number"]}',
                leave=False,
                unit='commit',
            )
        ],
    }


def get_pull_requests(
    client, api_repos, strip_text_content, pull_since, pull_until, redact_names_and_urls
):

    for repo in api_repos:
        try:
            for pr in tqdm(
                client.get_pullrequests(repo['full_name']),
                desc=f'downloading PRs for {repo["name"]}',
                unit='pr',
            ):
                updated_at = parser.parse(pr['updated_at'])

                # PRs are ordered newest to oldest; if this isn't old enough, skip it and keep going
                if updated_at >= pull_until:
                    continue

                # if this is too old, we're done with this repo
                if pull_since and updated_at < pull_since:
                    break

                yield _normalize_pr(client, pr, strip_text_content, redact_names_and_urls)

        except GithubException as e:
            print(f':WARN: Exception getting PRs for repo {repo["name"]}: {e}. Skipping...')
    print()



import re


class NameRedactor:
    def __init__(self, preserve_names=None):
        self.redacted_names = {}
        self.seq = 0
        self.preserve_names = preserve_names or []

    def redact_name(self, name):
        if name in self.preserve_names:
            return name

        redacted_name = self.redacted_names.get(name)
        if not redacted_name:
            redacted_name = f'redacted-{self.seq:04}'
            self.seq += 1
            self.redacted_names[name] = redacted_name
        return redacted_name


JIRA_KEY_REGEX = re.compile(r'([a-z0-9]+)[-|_|/| ]?(\d+)', re.IGNORECASE)


def sanitize_text(text, strip_text_content):
    if not text or not strip_text_content:
        return text

    return (' ').join(
        {f'{m[0].upper().strip()}-{m[1].upper().strip()}' for m in JIRA_KEY_REGEX.findall(text)}
    )
